from typing import Any, Dict
from unittest.mock import patch

import pytest

from ta_cmi import RateLimitError
from ta_cmi.api import ApiError
from ta_cmi.cmi_api import CMIAPI

from . import MAKE_GET_REQUEST_PACKAGE

TEST_HOST = "http://localhost"
TEST_USER = "user"
TEST_PASS = "pass"
TEST_NODE_ID = "1"
TEST_PARAMS = "I,O"


def generate_response(status_code: int) -> Dict[str, Any]:
    """Generate a response with the provided status code."""
    return {
        "Header": {"Version": 6, "Device": "8F", "Timestamp": 1670960961},
        "Data": {"DL-Bus": []},
        "Status": "OK",
        "Status code": status_code,
    }


def generate_device_data_url() -> str:
    """Generate the device data url."""
    return (
        f"{TEST_HOST}/INCLUDE/api.cgi?jsonparam={TEST_PARAMS}&jsonnode={TEST_NODE_ID}"
    )


api = CMIAPI(TEST_HOST, TEST_USER, TEST_PASS)


@pytest.mark.asyncio
async def test_get_devices_ids_to_list_conversion():
    """Test the get_devices_ids method with a valid api response."""
    with patch(
        "ta_cmi.api.API._make_request_no_json", return_value="1;2;"
    ) as request_mock:
        ids = await api.get_devices_ids()

        assert ids == ["1", "2"]

        request_mock.assert_called_once()


@pytest.mark.asyncio
async def test_get_device_data_successful_request():
    """Test the get_device_data function with a successful request."""
    sample_response: Dict[str, Any] = generate_response(0)

    with patch(MAKE_GET_REQUEST_PACKAGE, return_value=sample_response) as request_mock:
        result = await api.get_device_data(TEST_NODE_ID, TEST_PARAMS)

        assert result == sample_response

        request_mock.assert_called_once_with(generate_device_data_url())


@pytest.mark.asyncio
async def test_get_device_data_node_not_available():
    """Test the get_device_data function with a successful request but the cmi status code is 1."""
    with patch(
        MAKE_GET_REQUEST_PACKAGE, return_value=generate_response(1)
    ) as request_mock:
        with pytest.raises(ApiError) as exc_info:
            await api.get_device_data(TEST_NODE_ID, TEST_PARAMS)

        assert str(exc_info.value) == "Node not available"
        request_mock.assert_called_once_with(generate_device_data_url())


@pytest.mark.asyncio
async def test_get_device_data_fail_can_bus():
    """Test the get_device_data function with a successful request but the cmi status code is 2."""
    with patch(
        MAKE_GET_REQUEST_PACKAGE, return_value=generate_response(2)
    ) as request_mock:
        with pytest.raises(ApiError) as exc_info:
            await api.get_device_data(TEST_NODE_ID, TEST_PARAMS)

        assert (
            str(exc_info.value)
            == "Failure during the CAN-request/parameter not available for this device"
        )
        request_mock.assert_called_once_with(generate_device_data_url())


@pytest.mark.asyncio
async def test_get_device_data_rate_limit():
    """Test the get_device_data function with a successful request but the cmi status code is 4."""
    with patch(
        MAKE_GET_REQUEST_PACKAGE, return_value=generate_response(4)
    ) as request_mock:
        with pytest.raises(RateLimitError) as exc_info:
            await api.get_device_data(TEST_NODE_ID, TEST_PARAMS)

        assert str(exc_info.value) == "Only one request per minute is permitted"
        request_mock.assert_called_once_with(generate_device_data_url())


@pytest.mark.asyncio
async def test_get_device_data_unsupported_device():
    """Test the get_device_data function with a successful request but the cmi status code is 5."""
    with patch(
        MAKE_GET_REQUEST_PACKAGE, return_value=generate_response(5)
    ) as request_mock:
        with pytest.raises(ApiError) as exc_info:
            await api.get_device_data(TEST_NODE_ID, TEST_PARAMS)

        assert str(exc_info.value) == "Device not supported"
        request_mock.assert_called_once_with(generate_device_data_url())


@pytest.mark.asyncio
async def test_get_device_data_busy_can_bus():
    """Test the get_device_data function with a successful request but the cmi status code is 7."""
    with patch(
        MAKE_GET_REQUEST_PACKAGE, return_value=generate_response(7)
    ) as request_mock:
        with pytest.raises(ApiError) as exc_info:
            await api.get_device_data(TEST_NODE_ID, TEST_PARAMS)

        assert str(exc_info.value) == "CAN Bus is busy"
        request_mock.assert_called_once_with(generate_device_data_url())


@pytest.mark.asyncio
async def test_get_device_data_unknown_status_code():
    """Test the get_device_data function with a successful request but the cmi status code is unknown."""
    with patch(
        MAKE_GET_REQUEST_PACKAGE, return_value=generate_response(55)
    ) as request_mock:
        with pytest.raises(ApiError) as exc_info:
            await api.get_device_data(TEST_NODE_ID, TEST_PARAMS)

        assert str(exc_info.value) == "Unknown error"
        request_mock.assert_called_once_with(generate_device_data_url())



from dataclasses import dataclass
from typing import Any, Dict

from aiohttp import ClientSession

from .api import API
from .coe_channel import CoEChannel
from .const import _LOGGER, ChannelMode


@dataclass
class CoEServerConfig:
    coe_version: int


class CoEAPI(API):
    """Class to perform API requests to the CoE Addon."""

    COE_VERSION = "/version"
    COE_CONFIG = "/config"
    COE_DATA = "/receive/{id}"
    COE_SEND_ANALOG = "/send/analog"
    COE_SEND_DIGITAL = "/send/digital"
    COE_SEND_ANALOG_V2 = "/send/v2/analog"
    COE_SEND_DIGITAL_V2 = "/send/v2/digital"

    DIGITAL_VALUES_PER_PAGE = 16
    ANALOG_VALUES_PER_PAGE = 4

    V2_MIN_CHANNEL_INDEX = 1

    def __init__(self, host: str, session: ClientSession = None) -> None:
        """Initialize."""
        super().__init__(session)

        self.host = host

    async def get_coe_data(self, can_id: int) -> Dict[str, Any] | None:
        """Get the CoE data."""
        url = f"{self.host}{self.COE_DATA.replace('{id}', str(can_id))}"

        _LOGGER.debug("Receive data from CoE server: %s", url)

        data = await self._make_request_get(url)

        _LOGGER.debug("Received data from CoE server: %s", data)

        if len(data) == 0:
            return None

        return data

    async def get_coe_version(self) -> str | None:
        """Get the version of the CoE server."""
        url = f"{self.host}{self.COE_VERSION}"

        _LOGGER.debug("Receive current version from CoE server: %s", url)

        data = await self._make_request_get(url)

        _LOGGER.debug("Received version from CoE server: %s", data)

        if len(data) == 0:
            return None

        return data.get("version", None)

    async def get_coe_server_config(self) -> CoEServerConfig:
        """Get the CoE config from the server."""
        url = f"{self.host}{self.COE_CONFIG}"

        _LOGGER.debug("Receive current config from CoE server: %s", url)

        data = await self._make_request_get(url)

        _LOGGER.debug("Received version from CoE server: %s", data)

        return CoEServerConfig(coe_version=data.get("coe_version", 0))

    @staticmethod
    def _check_channel_mode(
        target_mode: ChannelMode, channel_to_check: list[CoEChannel]
    ) -> bool:
        """Check if the channel type equals the target."""
        for channel in channel_to_check:
            if channel.mode != target_mode:
                _LOGGER.warning(
                    f"Channel has wrong mode. Expected mode: {target_mode}, actual mode: {channel.mode}"
                )
                return False

        return True

    @staticmethod
    def _check_array_length(array: list, target_size: int) -> bool:
        """Check if a list has the required length."""
        if len(array) != target_size:
            _LOGGER.warning(
                f"List has wrong length. Expected length: {target_size}, actual length: {len(array)}"
            )
            return False

        return True

    @staticmethod
    def _check_analog_page_size(provided: int) -> bool:
        """Check if the page is in the right number range."""
        if not (0 < provided < 9):
            _LOGGER.warning(
                f"Page is not in the expected range. Expected range: 0 < page < 9, actual value: {provided}"
            )
            return False

        return True

    @staticmethod
    def _convert_analog_channel_to_dict(channel: CoEChannel) -> dict[str, Any]:
        """Convert a analog coe channel to a dict."""
        return {"value": channel.value, "unit": int(channel.unit)}

    async def send_analog_values(self, channels: list[CoEChannel], page: int):
        """Send analog values to the CoE server."""
        _LOGGER.debug("Send analog values to CoE server")

        if (
            not self._check_channel_mode(ChannelMode.ANALOG, channels)
            or not self._check_array_length(channels, self.ANALOG_VALUES_PER_PAGE)
            or not self._check_analog_page_size(page)
        ):
            _LOGGER.error("Could not send analog values. Please see logs for details.")
            return

        data = {
            "values": [
                self._convert_analog_channel_to_dict(channel) for channel in channels
            ],
            "page": page,
        }

        url = f"{self.host}{self.COE_SEND_ANALOG}"

        await self._make_request_post(url, data)

    async def send_digital_values(
        self, channels: list[CoEChannel], second_page: bool = False
    ) -> None:
        """Send digital values to the CoE server."""
        _LOGGER.debug("Send digital values to CoE server")

        if not self._check_channel_mode(
            ChannelMode.DIGITAL, channels
        ) or not self._check_array_length(channels, self.DIGITAL_VALUES_PER_PAGE):
            _LOGGER.error("Could not send digital values. Please see logs for details.")
            return

        data = {"values": [bool(x.value) for x in channels], "page": int(second_page)}

        url = f"{self.host}{self.COE_SEND_DIGITAL}"
        await self._make_request_post(url, data)

    @staticmethod
    def _map_channels_to_v2_body(channels: list[CoEChannel]) -> dict[str, Any]:
        """Map a channels to a v2 request body."""
        return {
            "values": [
                {"channel": channel.index, "value": channel.value, "unit": channel.unit}
                for channel in channels
            ],
        }

    @staticmethod
    def _check_channels_min_index(channels: list[CoEChannel], min_index: int) -> bool:
        """Check if the channel indexes are greate then the minimum."""
        for channel in channels:
            if channel.index < min_index:
                _LOGGER.warning(
                    f"Channel index must be greater than zero. Channel index: {channel.index}"
                )
                return False
        return True

    async def send_analog_values_v2(self, channel: list[CoEChannel]) -> None:
        """Send analog values to the CoE server (Version 2)."""
        _LOGGER.debug("Send analog value to CoE server (V2)")

        if not self._check_channel_mode(
            ChannelMode.ANALOG, channel
        ) or not self._check_channels_min_index(channel, self.V2_MIN_CHANNEL_INDEX):
            _LOGGER.error("Could not send analog values. Please see logs for details.")
            return

        data = self._map_channels_to_v2_body(channel)

        url = f"{self.host}{self.COE_SEND_ANALOG_V2}"
        await self._make_request_post(url, data)

    async def send_digital_values_v2(self, channel: list[CoEChannel]) -> None:
        """Send digital values to the CoE server (Version 2)."""
        _LOGGER.debug("Send digital value to CoE server (V2)")

        if not self._check_channel_mode(
            ChannelMode.DIGITAL, channel
        ) or not self._check_channels_min_index(channel, self.V2_MIN_CHANNEL_INDEX):
            _LOGGER.error("Could not send digital values. Please see logs for details.")
            return

        data = self._map_channels_to_v2_body(channel)

        url = f"{self.host}{self.COE_SEND_DIGITAL_V2}"
        await self._make_request_post(url, data)



import json
from typing import Any, Dict

import yarl
from aiohttp import BasicAuth, ClientConnectionError, ClientSession

from .const import _LOGGER, HTTP_UNAUTHORIZED


class API:
    """Class to perform basic API requests."""

    def __init__(self, session: ClientSession = None, auth: BasicAuth = None) -> None:
        """Initialize."""
        self.session = session
        self._auth = auth

        self._internal_session = False

    @staticmethod
    def parse_json(data: str) -> Dict[str, Any]:
        if not len(data):
            return {}

        return json.loads(data)

    async def _make_request_get(self, url: str) -> Dict[str, Any]:
        """Retrieve json data from REST GET endpoint."""
        return self.parse_json(await self._make_request_no_json(url))

    async def _make_request_post(
        self, url: str, body: dict[str, Any]
    ) -> Dict[str, Any]:
        """Retrieve json data from REST POST endpoint."""
        return self.parse_json(
            await self._make_request_no_json(url, method="POST", body=body)
        )

    async def _make_request_no_json(
        self, url: str, method: str = "GET", body: dict[str, Any] | None = None
    ) -> str:
        """Retrieve data from REST endpoint."""
        if self.session is None:
            self._internal_session = True
            self.session = ClientSession()

        try:
            _LOGGER.debug(f"Sending request to {url} with {body}")
            async with self.session.request(
                method, url, auth=self._auth, json=body
            ) as res:
                text = await res.text()
                await self._close_session()
                if res.status == HTTP_UNAUTHORIZED:
                    raise InvalidCredentialsError("Invalid credentials")
                elif res.status >= 300:
                    raise ApiError(
                        f"Invalid response from {res.url.host}: {res.status}"
                    )

                _LOGGER.debug(f"Received payload: {text}")
                return text
        except ClientConnectionError:
            await self._close_session()
            raise ApiError(f"Could not connect to {yarl.URL(url).host}")

    async def _close_session(self) -> None:
        """Close the internal session."""
        if self._internal_session:
            await self.session.close()
            self.session = None


class ApiError(Exception):
    """Raised when API request ended in error."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status


class InvalidCredentialsError(Exception):
    """Triggered when the credentials are invalid."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status



import json
from typing import Any, Dict, List

from aiohttp import BasicAuth, ClientSession

from .api import API, ApiError
from .const import _LOGGER


class CMIAPI(API):
    """Class to perform CMI API requests."""

    DEVICE_LIST = "/INCLUDE/can_nodes.cgi?_=1"
    DEVICE_DATA = "/INCLUDE/api.cgi?jsonparam=PARAMS&jsonnode=ID"

    def __init__(
        self, host: str, username: str, password: str, session: ClientSession = None
    ) -> None:
        """Initialize."""
        super().__init__(session, BasicAuth(username, password))

        self.host = host

    async def get_devices_ids(self) -> List[str]:
        """Get all id from devices connected to the CMI"""
        _LOGGER.debug("Receive list of nodes from C.M.I")

        url: str = f"{self.host}{self.DEVICE_LIST}"
        data: str = await self._make_request_no_json(url)

        _LOGGER.debug("Received list of nodes from C.M.I: %s", data)

        raw_ids: List[str] = data.split(";")

        return [x for x in raw_ids if len(x) != 0]

    async def get_device_data(self, node_id: str, parameter: str) -> Dict[str, Any]:
        """Get data from device."""
        _LOGGER.debug(
            "Make request to device %s with parameters: %s",
            node_id,
            parameter,
        )

        url: str = f"{self.host}{self.DEVICE_DATA.replace('PARAMS', parameter).replace('ID', node_id)}"

        return await self._make_data_request(url)

    async def _make_data_request(self, url: str) -> Dict[str, Any]:
        """Retrieve data from CMI API."""
        data = await self._make_request_get(url)

        STATUS_CODE = "Status code"

        if data[STATUS_CODE] != 0:
            _LOGGER.debug("Failed response received: %s", json.dumps(data))

        if data[STATUS_CODE] == 0:
            return data
        elif data[STATUS_CODE] == 1:
            raise ApiError("Node not available")
        elif data[STATUS_CODE] == 2:
            raise ApiError(
                "Failure during the CAN-request/parameter not available for this device"
            )
        elif data[STATUS_CODE] == 4:
            raise RateLimitError("Only one request per minute is permitted")
        elif data[STATUS_CODE] == 5:
            raise ApiError("Device not supported")
        elif data[STATUS_CODE] == 7:
            raise ApiError("CAN Bus is busy")
        else:
            raise ApiError("Unknown error")


class RateLimitError(Exception):
    """Triggered when the rate limit is reached."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status



import pytest

from ta_cmi.cmi_channel import CMIChannel
from ta_cmi.const import UNITS_DE, UNITS_EN, ChannelType, Languages


@pytest.mark.parametrize("unit_number, expected_unit", UNITS_DE.items())
def test_units_german(unit_number: str, expected_unit: str):
    """Test the translation from a unit number to a string with german units."""
    test = {"Number": 1, "AD": "A", "Value": {"Value": 0, "Unit": unit_number}}

    channel: CMIChannel = CMIChannel(ChannelType.INPUT, test)

    assert channel.get_unit(Languages.DE) == expected_unit


@pytest.mark.parametrize("unit_number, expected_unit", UNITS_EN.items())
def test_units_english(unit_number: str, expected_unit: str):
    """Test the translation from a unit number to a string with english units."""
    test = {"Number": 1, "AD": "A", "Value": {"Value": 0, "Unit": unit_number}}

    channel: CMIChannel = CMIChannel(ChannelType.INPUT, test)

    assert channel.get_unit(Languages.EN) == expected_unit



from copy import deepcopy
from unittest.mock import patch

import pytest

from ta_cmi import ChannelType, Device, InvalidDeviceError
from ta_cmi.cmi_api import CMIAPI
from ta_cmi.cmi_channel import CMIChannel
from ta_cmi.const import DEVICES, SUPPORTED_PARAMS_FOR_DEVICE

from . import DEVICE_DATA_PACKAGE, sleep_mock

TEST_NODE_ID = "1"
TEST_HOST = "http://localhost"

TEST_API = CMIAPI(TEST_HOST, "user", "pass")

DUMMY_RESPONSE = {
    "Header": {"Version": 5, "Device": "87", "Timestamp": 1630764000},
    "Data": {
        "Inputs": [
            {"Number": 1, "AD": "A", "Value": {"Value": 92.2, "Unit": "1"}},
        ],
        "Outputs": [
            {"Number": 1, "AD": "D", "Value": {"Value": 1, "Unit": "43"}},
        ],
        "Logging Analog": [
            {"Number": 1, "AD": "A", "Value": {"Value": 12.2, "Unit": "1"}},
        ],
        "Logging Digital": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "DL-Bus": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "General": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "Date": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "Time": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "Sun": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "Electrical power": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "Modbus": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "Network Analog": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "Network Digital": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "MBus": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
        "KNX": [
            {"Number": 1, "AD": "D", "Value": {"Value": 0, "Unit": "43"}},
        ],
    },
    "Status": "OK",
    "Status code": 0,
}


@pytest.mark.asyncio
async def test_device_update_no_type():
    """Test the update function with no device type set."""

    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    with patch(DEVICE_DATA_PACKAGE, return_value=DUMMY_RESPONSE) as device_data_mock:
        await device.update()

    # Request with default params
    device_data_mock.assert_called_once_with(TEST_NODE_ID, "I,O")

    assert device.device_id == DUMMY_RESPONSE["Header"]["Device"]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "type_id, expected_params", SUPPORTED_PARAMS_FOR_DEVICE.items()
)
async def test_device_update_type_set(type_id, expected_params):
    """Test the update function with a device type set."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    device.device_id = type_id

    with patch(DEVICE_DATA_PACKAGE, return_value=DUMMY_RESPONSE) as device_data_mock:
        await device.update()

    if len(expected_params.split(",")) > 7:
        assert device_data_mock.call_count == 2
        return

    # Request with default params
    device_data_mock.assert_called_once_with(TEST_NODE_ID, expected_params)


@pytest.mark.asyncio
async def test_device_with_more_than_seven_parameters():
    TARGET_DELAY = 61

    async def custom_sleep_mock(delay: int):
        assert delay == TARGET_DELAY

    device = Device(TEST_NODE_ID, TEST_API, custom_sleep_mock, TARGET_DELAY)

    device.set_device_type("UVR16x2")

    DUMMY_RESPONSE_PART1 = deepcopy(DUMMY_RESPONSE)
    DUMMY_RESPONSE_PART2 = deepcopy(DUMMY_RESPONSE)

    DUMMY_RESPONSE_PART1["Data"]["Inputs"] = {}
    DUMMY_RESPONSE_PART1["Data"]["Outputs"] = {}

    with patch(
        DEVICE_DATA_PACKAGE, side_effect=[DUMMY_RESPONSE_PART1, DUMMY_RESPONSE_PART2]
    ) as device_data_mock:
        await device.update()

    calls = device_data_mock.mock_calls

    assert "I,O,D,Sg,Sd,St,Ss" in calls[0].args

    assert "La,Ld" in calls[1].args

    assert device.has_channel_type(ChannelType.INPUT)
    assert device.has_channel_type(ChannelType.OUTPUT)
    assert device.has_channel_type(ChannelType.ANALOG_LOGGING)


@pytest.mark.asyncio
async def test_device_update_data_parse():
    """Test the parsing of the received data."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    with patch(DEVICE_DATA_PACKAGE, return_value=DUMMY_RESPONSE):
        await device.update()

    assert device.get_channels(ChannelType.INPUT) == {
        1: CMIChannel(ChannelType.INPUT, DUMMY_RESPONSE["Data"]["Inputs"][0])
    }
    assert device.get_channels(ChannelType.OUTPUT) == {
        1: CMIChannel(ChannelType.OUTPUT, DUMMY_RESPONSE["Data"]["Outputs"][0])
    }
    assert device.get_channels(ChannelType.ANALOG_LOGGING) == {
        1: CMIChannel(
            ChannelType.ANALOG_LOGGING, DUMMY_RESPONSE["Data"]["Logging Analog"][0]
        )
    }
    assert device.get_channels(ChannelType.DIGITAL_LOGGING) == {
        1: CMIChannel(
            ChannelType.DIGITAL_LOGGING, DUMMY_RESPONSE["Data"]["Logging Digital"][0]
        )
    }
    assert device.get_channels(ChannelType.DL_BUS) == {
        1: CMIChannel(ChannelType.DL_BUS, DUMMY_RESPONSE["Data"]["DL-Bus"][0])
    }

    assert device.get_channels(ChannelType.SYSTEM_VALUES_GENERAL) == {
        1: CMIChannel(
            ChannelType.SYSTEM_VALUES_GENERAL, DUMMY_RESPONSE["Data"]["General"][0]
        )
    }

    assert device.get_channels(ChannelType.SYSTEM_VALUES_TIME) == {
        1: CMIChannel(ChannelType.SYSTEM_VALUES_TIME, DUMMY_RESPONSE["Data"]["Time"][0])
    }

    assert device.get_channels(ChannelType.SYSTEM_VALUES_DATE) == {
        1: CMIChannel(ChannelType.SYSTEM_VALUES_DATE, DUMMY_RESPONSE["Data"]["Date"][0])
    }

    assert device.get_channels(ChannelType.SYSTEM_VALUES_SUN) == {
        1: CMIChannel(ChannelType.SYSTEM_VALUES_SUN, DUMMY_RESPONSE["Data"]["Sun"][0])
    }

    assert device.get_channels(ChannelType.SYSTEM_VALUES_E_POWER) == {
        1: CMIChannel(
            ChannelType.SYSTEM_VALUES_E_POWER,
            DUMMY_RESPONSE["Data"]["Electrical power"][0],
        )
    }

    assert device.get_channels(ChannelType.MODBUS) == {
        1: CMIChannel(ChannelType.MODBUS, DUMMY_RESPONSE["Data"]["Modbus"][0])
    }

    assert device.get_channels(ChannelType.NETWORK_ANALOG) == {
        1: CMIChannel(
            ChannelType.NETWORK_ANALOG, DUMMY_RESPONSE["Data"]["Network Analog"][0]
        )
    }

    assert device.get_channels(ChannelType.NETWORK_DIGITAL) == {
        1: CMIChannel(
            ChannelType.NETWORK_DIGITAL, DUMMY_RESPONSE["Data"]["Network Digital"][0]
        )
    }

    assert device.get_channels(ChannelType.MBUS) == {
        1: CMIChannel(ChannelType.MBUS, DUMMY_RESPONSE["Data"]["MBus"][0])
    }

    assert device.get_channels(ChannelType.KNX) == {
        1: CMIChannel(ChannelType.KNX, DUMMY_RESPONSE["Data"]["KNX"][0])
    }


@pytest.mark.asyncio
async def test_device_update_no_device_type():
    """Test the update function when no device type is given."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    with patch(DEVICE_DATA_PACKAGE, return_value=DUMMY_RESPONSE) as device_data_mock:
        await device.update()

    device_data_mock.assert_called_once_with(TEST_NODE_ID, "I,O")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "type_id",
    {k: DEVICES[k] for k in set(DEVICES) - set(SUPPORTED_PARAMS_FOR_DEVICE)}.keys(),
)
async def test_device_update_default_params(type_id):
    """Test the update function with a device that supports only the default params."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    device.device_id = type_id

    with patch(DEVICE_DATA_PACKAGE, return_value=DUMMY_RESPONSE) as device_data_mock:
        await device.update()

    device_data_mock.assert_called_once_with(TEST_NODE_ID, "I,O")


@pytest.mark.parametrize("type_id, expected_type_name", DEVICES.items())
def test_get_device_type_with_known_types(type_id, expected_type_name):
    """Test the get device function with types that are known."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    device.device_id = type_id

    assert device.get_device_type() == expected_type_name


def test_get_device_type_with_unknown_type():
    """Test the get device function with an unknown type."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    assert device.get_device_type() == "Unknown"


@pytest.mark.parametrize(
    "type_name, expected_type_id", [(value, key) for key, value in DEVICES.items()]
)
def test_set_device_type_with_valid_type(type_name: str, expected_type_id: str):
    """Test the set device type function with a valid type."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    assert device.device_id == "00"

    device.set_device_type(type_name)

    assert device.device_id == expected_type_id


def test_set_device_type_with_invalid_type():
    """Test the set device type function with an invalid type."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    with pytest.raises(InvalidDeviceError) as exc_info:
        device.set_device_type("NoValidType")

    assert str(exc_info.value) == "Invalid device name: NoValidType"


@pytest.mark.asyncio
@pytest.mark.parametrize("type_id, expected_type_name", DEVICES.items())
async def test_fetch_type_valid_type(type_id, expected_type_name):
    """Test the fetch type function with a valid type."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    assert device.get_device_type() == "Unknown"

    response = deepcopy(DUMMY_RESPONSE)
    response["Header"]["Device"] = type_id

    with patch(DEVICE_DATA_PACKAGE, return_value=response) as device_data_mock:
        await device.fetch_type()

    device_data_mock.assert_called_once_with(TEST_NODE_ID, "I,O")

    assert device.get_device_type() == expected_type_name


@pytest.mark.asyncio
async def test_fetch_type_invalid_type():
    """Test the fetch type function with an invalid type."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    assert device.get_device_type() == "Unknown"

    response = deepcopy(DUMMY_RESPONSE)
    response["Header"]["Device"] = "00"

    with patch(
        DEVICE_DATA_PACKAGE, return_value=response
    ) as device_data_mock, pytest.raises(InvalidDeviceError) as exc_info:
        await device.fetch_type()

    device_data_mock.assert_called_once_with(TEST_NODE_ID, "I,O")

    assert str(exc_info.value) == "Invalid device id: 00"


@pytest.mark.asyncio
@pytest.mark.parametrize("channel_type", list(ChannelType._member_map_.values()))
async def test_get_channels(channel_type):
    """Test the get channels function with all channel types."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    with patch(DEVICE_DATA_PACKAGE, return_value=DUMMY_RESPONSE):
        await device.update()

    channels = device.get_channels(channel_type)

    for ch in channels.values():
        assert ch.type == channel_type


def test_has_channel_type_exists():
    """Test the has channel type function with a type that was fetched."""
    device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    device._channels = {ChannelType.INPUT: {}}

    assert device.has_channel_type(ChannelType.INPUT)


def test_has_channel_type_non_exists():
    """Test the has channel type function with a type that was fetched."""
    device: Device = Device(TEST_NODE_ID, TEST_API, sleep_mock)

    device._channels = {ChannelType.INPUT: {}}

    assert not device.has_channel_type(ChannelType.OUTPUT)



__version__ = "3.4.0"
__author__ = "DeerMaximum"

from .api import API as API
from .api import ApiError as ApiError
from .api import InvalidCredentialsError as InvalidCredentialsError
from .channel import Channel as Channel
from .cmi import CMI as CMI
from .cmi_api import CMIAPI as CMIAPI
from .cmi_api import RateLimitError as RateLimitError
from .cmi_channel import CMIChannel as CMIChannel
from .coe import CoE as CoE
from .coe_api import CoEAPI as CoEAPI
from .coe_api import CoEServerConfig as CoEServerConfig
from .coe_channel import CoEChannel as CoEChannel
from .const import ChannelMode as ChannelMode
from .const import ChannelType as ChannelType
from .const import Languages as Languages
from .device import Device as Device
from .device import InvalidDeviceError as InvalidDeviceError



import asyncio
from typing import List

from aiohttp import ClientSession

from .cmi_api import CMIAPI
from .const import SLEEP_FUNCTION_TYPE, ReadOnlyClass
from .device import Device


class CMI(metaclass=ReadOnlyClass):
    """Main class to interact with CMI."""

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        session: ClientSession = None,
        sleep_function: SLEEP_FUNCTION_TYPE = asyncio.sleep,
        rate_limit_wait_time: int = 75,
    ) -> None:
        """Initialize."""
        self._api = CMIAPI(host, username, password, session)
        self._sleep_function = sleep_function
        self._rate_limit_wait_time = rate_limit_wait_time

    async def get_devices(self) -> List[Device]:
        """List connected devices."""
        device_ids = await self._api.get_devices_ids()

        return [
            Device(
                x,
                self._api,
                sleep_function=self._sleep_function,
                rate_limit_wait_time=self._rate_limit_wait_time,
            )
            for x in device_ids
        ]
